namespace eval StyleEditor {
	variable Info
	array unset Info
}

proc StyleEditor::Info {info args} {

    variable Info

    if {[llength $args]} {
	set Info($info) [lindex $args 0]
    }
    return $Info($info)
}

proc StyleEditor::Init {Tdemo} {

    set w .styleEditor
    toplevel $w
    wm title $w "TkTreeCtrl Style Editor"

    Info Tdemo $Tdemo

    panedwindow $w.pwH -orient horizontal -borderwidth 0

    panedwindow $w.pwH.pwV -orient vertical -borderwidth 0

    TreePlusScrollbarsInAFrame $w.pwH.pwV.styleList 0 1
    set T $w.pwH.pwV.styleList.t
    $T configure -showbuttons no -showlines no -showroot no -width 100 -height 200
    $T column create -text "Styles" -expand yes -button no -tag C0
    $T configure -treecolumn C0

    $T notify bind $T <Selection> {
	StyleEditor::SelectStyle
    }

    Info styleList $T

    TreePlusScrollbarsInAFrame $w.pwH.pwV.elementList 0 1
    set T $w.pwH.pwV.elementList.t
    $T configure -showbuttons no -showlines no -showroot no -width 100 -height 200
    $T column create -text "Elements" -expand yes -button no -tag C0
    $T configure -treecolumn C0

    $T notify bind $T <Selection> {
	StyleEditor::SelectElement
    }

    Info elementList $T

    $w.pwH.pwV add $w.pwH.pwV.styleList $w.pwH.pwV.elementList

    #
    # -orient
    #

    set f1 [frame $w.pwH.f1 -borderwidth 0]

    label $f1.labelStyle -textvariable ::StyleEditor::Info(labelStyle) -font {Helvetica 16}
    set lf [labelframe $f1.orient -text "-orient"]
    radiobutton $lf.h -text "horizontal" -variable ::StyleEditor::Info(-orient) \
	-value horizontal -command {StyleEditor::Sync_orient}
    radiobutton $lf.v -text "vertical" -variable ::StyleEditor::Info(-orient) \
	-value vertical -command {StyleEditor::Sync_orient}
    pack $lf.h -side top -anchor w
    pack $lf.v -side top -anchor w
    grid $f1.labelStyle -row 0 -column 0
    grid $lf -row 1 -column 0

    label $f1.labelElement -textvariable ::StyleEditor::Info(labelElement) -font {Helvetica 16}
    grid $f1.labelElement -row 2 -column 0

    #
    # -padx etc
    #

    frame $f1.fpad -borderwidth 0
    set row 0
    foreach option {-padx -pady -ipadx -ipady} {
	set ed [MakePadEditor $f1.fpad $option]
	grid $ed -row $row -column 0
	incr row
    }
    grid $f1.fpad -row 3 -column 0

    #
    # -expand etc
    #

    frame $f1.fexpand -borderwidth 0
    set row 0
    foreach option {-expand -iexpand} {
	set ed [MakeExpandEditor $f1.fexpand $option]
	grid $ed -row $row -column 0
	incr row
    }
    grid $f1.fexpand -row 3 -column 1

    #
    # -squeeze
    #

    set lf [labelframe $f1.squeeze -text "-squeeze"]
    foreach flag {x y} {
	checkbutton $lf.$flag -text $flag \
	    -variable ::StyleEditor::Info(-squeeze,$flag) \
	    -command {StyleEditor::Sync_squeeze}
	pack $lf.$flag -side top -anchor w
    }
    grid $lf -row 4 -column 1

    $w.pwH add $w.pwH.pwV $w.pwH.f1

    grid rowconfigure $w 0 -weight 1
    grid columnconfigure $w 0 -weight 1
    grid $w.pwH -row 0 -column 0 -sticky news

    return
}

proc StyleEditor::SetListOfStyles {} {

    set T [Info styleList]
    set Tdemo [Info Tdemo]

    # Create elements and styles the first time this is called
    if {[llength [$T style names]] == 0} {
	$T element create e1 text -fill [list $::SystemHighlightText {selected focus}]
	$T element create e2 rect -fill [list $::SystemHighlight {selected focus} gray {selected !focus}] \
	    -showfocus yes

	$T style create s1
	$T style elements s1 {e2 e1}
	$T style layout s1 e2 -union [list e1] -ipadx 2 -ipady 2 -iexpand e

	$T configure -defaultstyle s1
    }

    # Clear the list
    $T item delete all

    # One item for each style in the demo list
    foreach style [lsort -dictionary [$Tdemo style names]] {
	set I [$T item create]
	$T item text $I C0 $style
	$T item lastchild root $I

	Info item2style,$I $style
    }

    return
}

proc StyleEditor::SelectStyle {} {

    set T [Info styleList]
    set Tdemo [Info Tdemo]

    set selection [$T selection get]
    if {![llength $selection]} {
	[Info elementList] item delete all
	Info selectedStyle ""
	Info labelStyle "Style <none>"
	return
    }

    set I [lindex $selection 0]
    set style [Info item2style,$I]
    Info selectedStyle $style
    SetListOfElements $style

    Info labelStyle "Style $style"
    Info -orient [$Tdemo style cget $style -orient]

    return
}

proc StyleEditor::SetListOfElements {style} {

    set T [Info elementList]
    set Tdemo [Info Tdemo]

    # Create elements and styles the first time this is called
    if {[llength [$T style names]] == 0} {
	$T element create e1 text -fill [list $::SystemHighlightText {selected focus}]
	$T element create e2 rect -fill [list $::SystemHighlight {selected focus} gray {selected !focus}] \
	    -showfocus yes

	$T style create s1
	$T style elements s1 {e2 e1}
	$T style layout s1 e2 -union [list e1] -ipadx 2 -ipady 2 -iexpand e

	$T configure -defaultstyle s1
    }

    # Clear the list
    $T item delete all

    # One item for each element in the style
    foreach element [lsort -dictionary [$Tdemo style elements $style]] {
	set I [$T item create]
	$T item text $I C0 $element
	$T item lastchild root $I

	Info item2element,$I $element
    }

    return
}

proc StyleEditor::SelectElement {} {

    set T [Info elementList]
    set Tdemo [Info Tdemo]
    set style [Info selectedStyle]

    set selection [$T selection get]
    if {![llength $selection]} {
	Info selectedElement ""
	Info labelElement "Element <none>"
	foreach option {-padx -pady -ipadx -ipady} {
	    Info $option,1 0
	    Info $option,2 0
	}
	return
    }

    set I [lindex $selection 0]
    set element [Info item2element,$I]
    Info selectedElement $element

    Info labelElement "Element $element"

    foreach option {-padx -pady -ipadx -ipady} {
	set pad [$Tdemo style layout $style $element $option]
	if {[llength $pad] == 2} {
	    Info $option,1 [lindex $pad 0]
	    Info $option,2 [lindex $pad 1]
	} else {
	    Info $option,1 $pad
	    Info $option,2 $pad
	}
    }

    foreach option {-expand -iexpand} {
	set value [$Tdemo style layout $style $element $option]
	foreach flag {n s w e} {
	    Info $option,$flag [expr {[string first $flag $value] != -1}]
	}
    }

    set value [$Tdemo style layout $style $element -squeeze]
    foreach flag {x y} {
	Info -squeeze,$flag [expr {[string first $flag $value] != -1}]
    }

    return
}

proc StyleEditor::MakePadEditor {parent option} {

    set f [labelframe $parent.$option -text $option]
    spinbox $f.v1 -from 0 -to 100 -width 3 \
	-command [list StyleEditor::Sync_pad $option] \
	-textvariable ::StyleEditor::Info($option,1)
    spinbox $f.v2 -from 0 -to 100 -width 3 \
	-command [list StyleEditor::Sync_pad $option] \
	-textvariable ::StyleEditor::Info($option,2)
    pack $f.v1 -side left -padx 10 -pady 5
    pack $f.v2 -side left -padx {0 10} -pady 5
    
    return $f
}

proc StyleEditor::MakeExpandEditor {parent option} {

    set f [labelframe $parent.$option -text $option]
    foreach flag {n s w e} {
	checkbutton $f.$flag -text $flag -width 1 \
	    -variable ::StyleEditor::Info($option,$flag) \
	    -command [list StyleEditor::Sync_expand $option]
    }
    grid $f.n -row 0 -column 1
    grid $f.s -row 2 -column 1
    grid $f.w -row 1 -column 0
    grid $f.e -row 1 -column 2

    return $f
}

proc StyleEditor::Sync_orient {} {

    set Tdemo [Info Tdemo]
    set style [Info selectedStyle]
    $Tdemo style configure $style -orient [Info -orient]
    return
}

proc StyleEditor::Sync_pad {option} {

    set Tdemo [Info Tdemo]
    set style [Info selectedStyle]
    set element [Info selectedElement]
    $Tdemo style layout $style $element $option [list [Info $option,1] [Info $option,2]]
    return
}

proc StyleEditor::Sync_expand {option} {

    set Tdemo [Info Tdemo]
    set style [Info selectedStyle]
    set element [Info selectedElement]
    set value ""
    foreach flag {n s w e} {
	if {[Info $option,$flag]} {
	    append value $flag
	}
    }
    $Tdemo style layout $style $element $option $value
    return
}

proc StyleEditor::Sync_squeeze {} {

    set Tdemo [Info Tdemo]
    set style [Info selectedStyle]
    set element [Info selectedElement]
    set value ""
    foreach flag {x y} {
	if {[Info -squeeze,$flag]} {
	    append value $flag
	}
    }
    $Tdemo style layout $style $element -squeeze $value
    return
}

if 0 {


proc StyleEditor:: {} {
    return
}

proc StyleEditor:: {} {
    return
}

proc StyleEditor:: {} {
    return
}

proc StyleEditor:: {} {
    return
}

proc StyleEditor:: {} {
    return
}

proc StyleEditor:: {} {
    return
}

proc StyleEditor:: {} {
    return
}

proc StyleEditor:: {} {
    return
}
}
